/*	$OpenBSD: rstat_proc.c,v 1.10 1998/07/10 08:06:10 deraadt Exp $	*/

/* 
 * Copyright (c) 2010, Oracle America, Inc.
 *
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * Neither the name of the “Oracle America, Inc.” nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef lint
/*static char sccsid[] = "from: @(#)rpc.rstatd.c 1.1 86/09/25 Copyr 1984 Sun Micro";*/
/*static char sccsid[] = "from: @(#)rstat_proc.c	2.2 88/08/01 4.0 RPCSRC";*/
static char rcsid[] = "$OpenBSD: rstat_proc.c,v 1.10 1998/07/10 08:06:10 deraadt Exp $";
#endif

/*
 * rstat service:  built with rstat.x and derived from rpc.rstatd.c
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <locale.h>

#include <limits.h>
#include <rpc/rpc.h>
#include <sys/socket.h>
#include <syslog.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <unistd.h>

#if defined(__linux__)

#include <assert.h>
#include <ctype.h>

#define	CP_USER		0
#define	CP_NICE		1
#define	CP_SYS		2
#define	CP_IDLE		3
#define	CP_INTR		4
#define	CPUSTATES	5
static int cp_xlat[CPUSTATES] = { CP_USER, CP_NICE, CP_SYS, CP_IDLE };

/*
 * System wide statistics counters.
 */
struct vmmeter {
	/*
	 * General system activity.
	 */
	u_int v_swtch;		/* context switches */
	u_int v_trap;		/* calls to trap */
	u_int v_syscall;	/* calls to syscall() */
	u_int v_intr;		/* device interrupts */
	u_int v_soft;		/* software interrupts */
	u_int v_faults;		/* total faults taken */
	/*
	 * Virtual memory activity.
	 */
	u_int v_lookups;	/* object cache lookups */
	u_int v_hits;		/* object cache hits */
	u_int v_vm_faults;	/* number of address memory faults */
	u_int v_cow_faults;	/* number of copy-on-writes */
	u_int v_swpin;		/* swapins */
	u_int v_swpout;		/* swapouts */
	u_int v_pswpin;		/* pages swapped in */
	u_int v_pswpout;	/* pages swapped out */
	u_int v_pageins;	/* number of pageins */
	u_int v_pageouts;	/* number of pageouts */
	u_int v_pgpgin;		/* pages paged in */
	u_int v_pgpgout;	/* pages paged out */
	u_int v_intrans;	/* intransit blocking page faults */
	u_int v_reactivated;	/* number of pages reactivated from free list */
	u_int v_rev;		/* revolutions of the hand */
	u_int v_scan;		/* scans in page out daemon */
	u_int v_dfree;		/* pages freed by daemon */
	u_int v_pfree;		/* pages freed by exiting processes */
	u_int v_zfod;		/* pages zero filled on demand */
	u_int v_nzfod;		/* number of zfod's created */
	/*
	 * Distribution of page usages.
	 */
	u_int v_page_size;	/* page size in bytes */
	u_int v_kernel_pages;	/* number of pages in use by kernel */
	u_int v_free_target;	/* number of pages desired free */
	u_int v_free_min;	/* minimum number of pages desired free */
	u_int v_free_count;	/* number of pages free */
	u_int v_wire_count;	/* number of pages wired down */
	u_int v_active_count;	/* number of pages active */
	u_int v_inactive_target; /* number of pages desired inactive */
	u_int v_inactive_count;  /* number of pages inactive */
};
static struct	vmmeter cnt;

/* poseur disk entry to hold the information we're interested in. */
struct _disk {
	int		 *dk_select;	/* Display stats for selected disks. */
	char		**dk_name;	/* Disk names (sd0, wd1, etc). */
	u_int64_t	 *dk_xfer;	/* # of transfers. */
	u_int64_t	 *dk_seek;	/* # of seeks (currently unused). */
	u_int64_t	 *dk_bytes;	/* # of bytes transfered. */
	struct timeval	 *dk_time;	/* Time spent in disk i/o. */
	long	tk_nin;			/* TTY Chars in. */
	long	tk_nout;		/* TTY Chars out. */
	long	cp_time[CPUSTATES];	/* System timer ticks. */
};

#define	DK_NDRIVE	4
static int dk_ndrive = 0;

static int		cdk_select[DK_NDRIVE];
static char *		cdk_name[DK_NDRIVE];
static u_int64_t	cdk_xfer[DK_NDRIVE];
static u_int64_t	cdk_seek[DK_NDRIVE];
static u_int64_t	cdk_bytes[DK_NDRIVE];
static struct timeval	cdk_time[DK_NDRIVE];
static struct _disk cur = {
	cdk_select, cdk_name, cdk_xfer, cdk_seek, cdk_bytes, cdk_time,
	0, 0,
	{0, 0, 0, 0, 0}
};

#if 0
static int		ldk_select[DK_NDRIVE];
static char *		ldk_name[DK_NDRIVE];
static u_int64_t	ldk_xfer[DK_NDRIVE];
static u_int64_t	ldk_seek[DK_NDRIVE];
static u_int64_t	ldk_bytes[DK_NDRIVE];
static struct timeval	ldk_time[DK_NDRIVE];
static struct _disk last = {
	ldk_select, ldk_name, ldk_xfer, ldk_seek, ldk_bytes, ldk_time,
	0, 0,
	{0, 0, 0, 0, 0}
};
#endif

#else	/* __linux__ */

#include <kvm.h>
#include <nlist.h>

#ifdef BSD
#include <sys/vmmeter.h>
#include <sys/dkstat.h>
#include "dkstats.h"
#else
#include <sys/dk.h>
#endif
#endif	/* __linux__ */

#include <net/if.h>

#undef FSHIFT			 /* Use protocol's shift and scale values */
#undef FSCALE
#undef DK_NDRIVE
#undef CPUSTATES
#undef if_ipackets
#undef if_ierrors
#undef if_opackets
#undef if_oerrors
#undef if_collisions
#include <rpcsvc/rstat.h>

#ifdef BSD
#define BSD_CPUSTATES	5	/* Use protocol's idea of CPU states */
int	cp_xlat[CPUSTATES] = { CP_USER, CP_NICE, CP_SYS, CP_IDLE };
#endif

#if defined(__linux__)
struct user_net_device_stats {
    unsigned long rx_packets;	/* total packets received       */
    unsigned long tx_packets;	/* total packets transmitted    */
    unsigned long rx_bytes;	/* total bytes received         */
    unsigned long tx_bytes;	/* total bytes transmitted      */
    unsigned long rx_errors;	/* bad packets received         */
    unsigned long tx_errors;	/* packet transmit problems     */
    unsigned long rx_dropped;	/* no space in linux buffers    */
    unsigned long tx_dropped;	/* no space available in linux  */
    unsigned long rx_multicast;	/* multicast packets received   */
    unsigned long rx_compressed;
    unsigned long tx_compressed;
    unsigned long collisions;

    /* detailed rx_errors: */
    unsigned long rx_length_errors;
    unsigned long rx_over_errors;	/* receiver ring buff overflow  */
    unsigned long rx_crc_errors;	/* recved pkt with crc error    */
    unsigned long rx_frame_errors;	/* recv'd frame alignment error */
    unsigned long rx_fifo_errors;	/* recv'r fifo overrun          */
    unsigned long rx_missed_errors;	/* receiver missed packet     */
    /* detailed tx_errors */
    unsigned long tx_aborted_errors;
    unsigned long tx_carrier_errors;
    unsigned long tx_fifo_errors;
    unsigned long tx_heartbeat_errors;
    unsigned long tx_window_errors;
};

struct interface {
    struct interface *next;

    char name[IFNAMSIZ];	/* interface name        */
    short type;			/* if type               */
    short flags;		/* various flags         */
    int metric;			/* routing metric        */
    int mtu;			/* MTU value             */
    int tx_queue_len;		/* transmit queue length */
    struct ifmap map;		/* hardware setup        */
    struct sockaddr addr;	/* IP address            */
    struct sockaddr dstaddr;	/* P-P IP address        */
    struct sockaddr broadaddr;	/* IP broadcast address  */
    struct sockaddr netmask;	/* IP network mask       */
    struct sockaddr ipxaddr_bb;	/* IPX network address   */
    struct sockaddr ipxaddr_sn;	/* IPX network address   */
    struct sockaddr ipxaddr_e3;	/* IPX network address   */
    struct sockaddr ipxaddr_e2;	/* IPX network address   */
    struct sockaddr ddpaddr;	/* Appletalk DDP address */
    struct sockaddr ecaddr;	/* Econet address        */
    int has_ip;
    int has_ipx_bb;
    int has_ipx_sn;
    int has_ipx_e3;
    int has_ipx_e2;
    int has_ax25;
    int has_ddp;
    int has_econet;
    char hwaddr[32];		/* HW address            */
    int statistics_valid;
    struct user_net_device_stats stats;		/* statistics            */
    int keepalive;		/* keepalive value for SLIP */
    int outfill;		/* outfill value for SLIP */
};

#else	/* __linux__ */

struct nlist nl[] = {
#define	X_CNT		0
	{ "_cnt" },
#define	X_IFNET		1
	{ "_ifnet" },
#define	X_BOOTTIME	2
	{ "_boottime" },
#ifndef BSD
#define	X_HZ		3
	{ "_hz" },
#define	X_CPTIME	4
	{ "_cp_time" },
#define	X_DKXFER	5
	{ "_dk_xfer" },
#endif
	{ NULL },
};

#ifdef BSD
extern int dk_ndrive;		/* from dkstats.c */
extern struct _disk cur, last;
char *memf = NULL, *nlistf = NULL;
#endif
int hz;

struct ifnet_head ifnetq;	/* chain of ethernet interfaces */
int numintfs;
int stats_service();
kvm_t *kfd;

#endif	/* __linux__ */

extern int from_inetd;
int sincelastreq = 0;		/* number of alarms since last request */
extern int closedown;

struct {
	struct stats s1;
	struct statsswtch s2;
	struct statstime s3;
} stats_all;

static void setup(void);
static void updatestat(int signo);
static int havedisk(void);
void	rstat_service(struct svc_req *rqstp, SVCXPRT *transp);

static int stat_is_init = 0;
extern int errno;

#ifndef FSCALE
#define FSCALE (1 << 8)
#endif

static void
stat_init(void)
{
	stat_is_init = 1;

	/* Prime the pump */
	updatestat(0);

	(void) signal(SIGALRM, updatestat);
	alarm(1);
	pause();
}

statstime *
rstatproc_stats_3_svc(void *arg, struct svc_req *rqstp)
{
	(void)&arg; (void)&rqstp;
	if (!stat_is_init)
		stat_init();
	sincelastreq = 0;
	return (&stats_all.s3);
}

statsswtch *
rstatproc_stats_2_svc(void *arg, struct svc_req *rqstp)
{
	(void)&arg; (void)&rqstp;
	if (!stat_is_init)
		stat_init();
	sincelastreq = 0;
	return (&stats_all.s2);
}

stats *
rstatproc_stats_1_svc(void *arg, struct svc_req *rqstp)
{
	(void)&arg; (void)&rqstp;
	if (!stat_is_init)
		stat_init();
	sincelastreq = 0;
	return (&stats_all.s1);
}

u_int *
rstatproc_havedisk_3_svc(void *arg, struct svc_req *rqstp)
{
	static u_int have;

	(void)&arg; (void)&rqstp;
	if (!stat_is_init)
		stat_init();
	have = havedisk();
	sincelastreq = 0;
	return (&have);
}

u_int *
rstatproc_havedisk_2_svc(void *arg, struct svc_req *rqstp)
{
	(void)&arg; (void)&rqstp;
	return (rstatproc_havedisk_3_svc(arg, rqstp));
}

u_int *
rstatproc_havedisk_1_svc(void *arg, struct svc_req *rqstp)
{
	(void)&arg; (void)&rqstp;
	return (rstatproc_havedisk_3_svc(arg, rqstp));
}

#if defined(__linux__)

#define MAX_DISKS 4

struct _ldisk {
	unsigned int xfer[MAX_DISKS];
	unsigned int rio[MAX_DISKS];
	unsigned int wio[MAX_DISKS];
	unsigned int rblk[MAX_DISKS];
	unsigned int wblk[MAX_DISKS];
};

static void
getdiskstat(struct _ldisk *d)
{
  FILE *statfp;
  int dsk = 0;
  unsigned int active;
  unsigned int ma, mi, xfer;
  /* merges, ticks, in flight meals, and aveq are not used */
  unsigned int rd_ios, rd_merges, rd_sectors, rd_ticks;
  unsigned int wr_ios, wr_merges, wr_sectors, wr_ticks;
  unsigned int ios_in_flight, io_ticks, aveq;
  int cd_check = 0;
  char disk_name[256];
  static char line[2560];
  char *s;
  
  /* Open /proc/partitions file */
  if ((statfp = fopen("/proc/diskstats", "r")) == NULL) {
    syslog(LOG_ERR, "Cannot open %s: %s\n", "/proc/diskstats", strerror(errno));
    exit(1);
  }

  /* The first two lines are unimportant. */
  fgets(line, 2560, statfp);
  fgets(line, 2560, statfp);
  while (fgets(line, 2560, statfp) != NULL && dsk < MAX_DISKS) {
    sscanf (line, "%4d %4d %s %d %d %d %d %d %d %d %d %d %d %d",
            &ma, &mi, disk_name,
            &rd_ios, &rd_merges, &rd_sectors, &rd_ticks,
            &wr_ios, &wr_merges, &wr_sectors, &wr_ticks,
            &ios_in_flight, &io_ticks, &aveq);
    /* Check for compaq devices */
    for(s=disk_name;*s;s++) {
      if(*s == '/') {
        cd_check=1;
      }
    }

    /*
     * if we set cd_check we need to check
     * the name of the device using the compaq c#d#p# 
     * format, else we can just use the h/sda# format
     */
    if(cd_check) {
      /*
       * if the second or third from the last character
       * is p then this is a partition
       */
      if((s[-2] == 'p') || (s[-3] == 'p'))
        continue;
      goto found_dsk;
    } else {
      /*
       * if the last value is a digit, its a
       * partition, so go back to the top of the whlie loop
       */
      if (isdigit(s[-1])) {
        continue;
      }
found_dsk:
      xfer = rd_ios + wr_ios;
      /* this is how /proc/stat determines which devices to display in /proc/stat */
      active = xfer + rd_sectors, wr_sectors;
      if (active) {
        dsk++;
        d->xfer[dsk] = xfer;
        d->rio[dsk] = rd_ios;
        d->rblk[dsk] = rd_sectors;
        d->wio[dsk] = wr_ios;
        d->wblk[dsk] = wr_sectors;
      } else {
        /* Do not count inactive devices. */
      }
    }
  }
  /* Close /proc/partitions file */
   fclose(statfp);
}

static void
getstat(unsigned long *cuse, unsigned long *cice, unsigned long *csys, unsigned long *cide,
	     unsigned *pin, unsigned *pout, unsigned *sin, unsigned *sout,
	     unsigned *itot, unsigned *i1, unsigned *ct)
{
  static int stat;
#define	BUFFSIZE	1024
  char *buff = NULL;
  off_t buffsize = 1;
  int ndisks;
  
  if ((stat=open("/proc/stat", O_RDONLY, 0)) != -1) {
    char* b;
    b = buff;

    do {
      buffsize += BUFFSIZE;
      buff = realloc(buff, buffsize);
      if (buff == NULL) {
	perror("getstat(): realloc");
	abort();
      }
      b = buff + buffsize - BUFFSIZE - 1;
    } while (read(stat,b,BUFFSIZE) == BUFFSIZE);
    close(stat);
    buff[buffsize - 1] = '\0';
    
    *itot = 0; 
    *i1 = 1;   /* ensure assert below will fail if the sscanf bombs */
    b = strstr(buff, "cpu ");
    if(b)
    sscanf(b, "cpu  %lu %lu %lu %lu", cuse, cice, csys, cide);
    b = strstr(buff, "page ");
    if(b)
    sscanf(b, "page %u %u", pin, pout);
    b = strstr(buff, "swap ");
    if(b)
    sscanf(b, "swap %u %u", sin, sout);
    b = strstr(buff, "intr ");
    if(b)
    sscanf(b, "intr %u %u", itot, i1);
    b = strstr(buff, "ctxt ");
    if(b)
    sscanf(b, "ctxt %u", ct);
    /* avoid breaking with certain kernels where all the
       interrupt numbers have 1000000000 added to the value */
    /*    assert(*itot>*i1); */
    free(buff);
  }
}

static int procnetdev_vsn;

static char *
get_name(char *name, char *p)
{
    while (isspace(*p))
	p++;
    while (*p) {
	if (isspace(*p))
	    break;
	if (*p == ':') {	/* could be an alias */
	    char *dot = p, *dotname = name;
	    *name++ = *p++;
	    while (isdigit(*p))
		*name++ = *p++;
	    if (*p != ':') {	/* it wasn't, backup */
		p = dot;
		name = dotname;
	    }
	    if (*p == '\0')
		return NULL;
	    p++;
	    break;
	}
	*name++ = *p++;
    }
    *name++ = '\0';
    return p;
}

static int procnetdev_version(char *buf)
{
    if (strstr(buf, "compressed"))
	return 3;
    if (strstr(buf, "bytes"))
	return 2;
    return 1;
}

static int
get_dev_fields(char *bp, struct interface *ife)
{
    switch (procnetdev_vsn) {
    case 3:
	if(bp)
	sscanf(bp,
	"%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
	       &ife->stats.rx_bytes,
	       &ife->stats.rx_packets,
	       &ife->stats.rx_errors,
	       &ife->stats.rx_dropped,
	       &ife->stats.rx_fifo_errors,
	       &ife->stats.rx_frame_errors,
	       &ife->stats.rx_compressed,
	       &ife->stats.rx_multicast,

	       &ife->stats.tx_bytes,
	       &ife->stats.tx_packets,
	       &ife->stats.tx_errors,
	       &ife->stats.tx_dropped,
	       &ife->stats.tx_fifo_errors,
	       &ife->stats.collisions,
	       &ife->stats.tx_carrier_errors,
	       &ife->stats.tx_compressed);
	break;
    case 2:
	if(bp)
	sscanf(bp, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
	       &ife->stats.rx_bytes,
	       &ife->stats.rx_packets,
	       &ife->stats.rx_errors,
	       &ife->stats.rx_dropped,
	       &ife->stats.rx_fifo_errors,
	       &ife->stats.rx_frame_errors,

	       &ife->stats.tx_bytes,
	       &ife->stats.tx_packets,
	       &ife->stats.tx_errors,
	       &ife->stats.tx_dropped,
	       &ife->stats.tx_fifo_errors,
	       &ife->stats.collisions,
	       &ife->stats.tx_carrier_errors);
	ife->stats.rx_multicast = 0;
	break;
    case 1:
	if(bp)
	sscanf(bp, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
	       &ife->stats.rx_packets,
	       &ife->stats.rx_errors,
	       &ife->stats.rx_dropped,
	       &ife->stats.rx_fifo_errors,
	       &ife->stats.rx_frame_errors,

	       &ife->stats.tx_packets,
	       &ife->stats.tx_errors,
	       &ife->stats.tx_dropped,
	       &ife->stats.tx_fifo_errors,
	       &ife->stats.collisions,
	       &ife->stats.tx_carrier_errors);
	ife->stats.rx_bytes = 0;
	ife->stats.tx_bytes = 0;
	ife->stats.rx_multicast = 0;
	break;
    }
    return 0;
}

static int hz;
static struct timeval tm, btm;
static double avrun[3];
static unsigned long cpu_use[2], cpu_nic[2], cpu_sys[2];
static unsigned long cpu_idl[2];
static struct _ldisk disk[2];
static unsigned int pgpgin[2], pgpgout[2], pswpin[2], pswpout[2];
static unsigned int inter[2], ticks[2], ctxt[2];

#ifdef NOTYET	/* XXX noop to match solaris behavior */
#define	_DELTA(_x, _i)	( (_x)[(_i)] - (_x)[!(_i)] )
#else
#define	_DELTA(_x, _i)	( (_x)[(_i)] )
#endif

static void
convert_stats(void)
{
	static int tog = 0;
	double up;
	int i;
	int fd;
	static char buf[1024];
	char *savelocale;

	hz = sysconf(_SC_CLK_TCK);
	gettimeofday(&tm, (struct timezone *) 0);

	fd = open("/proc/uptime", O_RDONLY);
	if (fd == -1) {
		fprintf(stderr, "Error opening /proc/uptime, please make sure /proc is mounted.\n");
		exit(1);
	}

	i = read(fd, buf, sizeof(buf) - 1);
	if (i < 0) {
		perror("/proc/uptime");
		close(fd);
		exit(1);
	}
	buf[i] = '\0';
	savelocale = setlocale(LC_NUMERIC, NULL);
	setlocale(LC_NUMERIC, "C");
	if (sscanf(buf, "%lf", &up) < 1) {
		fprintf(stderr, "bad data in /proc/uptime\n");
		exit(1);
	}
	setlocale(LC_NUMERIC, savelocale);
	close(fd);

	i = (int)up;

	btm.tv_sec = tm.tv_sec - i;
	btm.tv_usec = 0;
	up -= i;
	if (up > 0)
		btm.tv_usec = 1000000 * up;

	fd = open("/proc/loadavg", O_RDONLY);
	if (fd == -1) {
		fprintf(stderr, "Error opening /proc/loadavg, please make sure /proc is mounted.\n");
		exit(1);
	}

	i = read(fd, buf, sizeof(buf) - 1);
	if (i < 0) {
		perror("/proc/loadavg");
		close(fd);
		exit(1);
	}
	buf[i] = '\0';
	savelocale = setlocale(LC_NUMERIC, NULL);
	setlocale(LC_NUMERIC, "C");
	if (sscanf(buf, "%lf %lf %lf", &avrun[0], &avrun[1], &avrun[2]) < 3) {
		fprintf(stderr, "bad data in /proc/loadavg\n");
		exit(1);
	}
	setlocale(LC_NUMERIC, savelocale);
	close(fd);

	getstat(cpu_use+tog,cpu_nic+tog,cpu_sys+tog,cpu_idl+tog,
		pgpgin+tog,pgpgout+tog,pswpin+tog,pswpout+tog,
		inter+tog,ticks+tog,ctxt+tog);
	getdiskstat(disk+tog);

	dk_ndrive = 0;
	for (i = 0; i < 4; i++) {
		if (disk[tog].xfer[i] == 0)
			continue;
		cur.dk_xfer[dk_ndrive] = disk[tog].xfer[i] - disk[!tog].xfer[i];
		dk_ndrive++;
	}

	cur.cp_time[CP_USER] = (_DELTA(cpu_use, tog)) % UINT_MAX;
	cur.cp_time[CP_NICE] = (_DELTA(cpu_nic, tog)) % UINT_MAX;
	cur.cp_time[CP_SYS]  = (_DELTA(cpu_sys, tog)) % UINT_MAX;
	cur.cp_time[CP_INTR] = 0;
	cur.cp_time[CP_IDLE] = (_DELTA(cpu_idl, tog)) % UINT_MAX;

	cnt.v_pgpgin = pgpgin[tog];
	cnt.v_pgpgout = pgpgout[tog];
	cnt.v_pswpin = pswpin[tog];
	cnt.v_pswpout = pswpout[tog];
	cnt.v_intr = inter[tog];
	cnt.v_swtch = ctxt[tog];
	tog = !tog;
}

static void
updatestat(int signo)
{
	int i, save_errno = errno;

#ifdef DEBUG
	syslog(LOG_DEBUG, "entering updatestat");
#endif
	if (signo != 0 && sincelastreq >= closedown) {
#ifdef DEBUG
		syslog(LOG_DEBUG, "about to closedown");
#endif
		if (from_inetd) {
			exit(0);
		} else {
			stat_is_init = 0;
			return;
		}
	}
	sincelastreq++;

/* ======================================= */
	convert_stats();

	memset(&stats_all, 0, sizeof(stats_all));

	for (i = 0; i < dk_ndrive && i < DK_NDRIVE; i++) {
		stats_all.s1.dk_xfer[i] = cur.dk_xfer[i];
		stats_all.s2.dk_xfer[i] = cur.dk_xfer[i];
		stats_all.s3.dk_xfer[i] = cur.dk_xfer[i];
        }

	for (i = 0; i < CPUSTATES; i++) {
		stats_all.s1.cp_time[i] += cur.cp_time[cp_xlat[i]];
		stats_all.s2.cp_time[i] += cur.cp_time[cp_xlat[i]];
		stats_all.s3.cp_time[i] += cur.cp_time[cp_xlat[i]];
        }

	stats_all.s2.avenrun[0] = avrun[0] * FSCALE;
	stats_all.s2.avenrun[1] = avrun[1] * FSCALE;
	stats_all.s2.avenrun[2] = avrun[2] * FSCALE;
	stats_all.s3.avenrun[0] = avrun[0] * FSCALE;
	stats_all.s3.avenrun[1] = avrun[1] * FSCALE;
	stats_all.s3.avenrun[2] = avrun[2] * FSCALE;

	stats_all.s2.boottime.tv_sec = btm.tv_sec;
	stats_all.s2.boottime.tv_usec = btm.tv_usec;
	stats_all.s3.boottime.tv_sec = btm.tv_sec;
	stats_all.s3.boottime.tv_usec = btm.tv_usec;

#ifdef DEBUG
	syslog(LOG_DEBUG, "%d %d %d %d", stats_all.s1.cp_time[0],
	    stats_all.s1.cp_time[1], stats_all.s1.cp_time[2],
	    stats_all.s1.cp_time[3]);
#endif

	stats_all.s1.v_pgpgin = cnt.v_pgpgin;
	stats_all.s1.v_pgpgout = cnt.v_pgpgout;
	stats_all.s1.v_pswpin = cnt.v_pswpin;
	stats_all.s1.v_pswpout = cnt.v_pswpout;
	stats_all.s1.v_intr = cnt.v_intr;
	stats_all.s2.v_pgpgin = cnt.v_pgpgin;
	stats_all.s2.v_pgpgout = cnt.v_pgpgout;
	stats_all.s2.v_pswpin = cnt.v_pswpin;
	stats_all.s2.v_pswpout = cnt.v_pswpout;
	stats_all.s2.v_intr = cnt.v_intr;
	stats_all.s3.v_pgpgin = cnt.v_pgpgin;
	stats_all.s3.v_pgpgout = cnt.v_pgpgout;
	stats_all.s3.v_pswpin = cnt.v_pswpin;
	stats_all.s3.v_pswpout = cnt.v_pswpout;
	stats_all.s3.v_intr = cnt.v_intr;
	gettimeofday(&tm, (struct timezone *) 0);
	stats_all.s1.v_intr -= hz*(tm.tv_sec - btm.tv_sec) +
	    hz*(tm.tv_usec - btm.tv_usec)/1000000;
	stats_all.s2.v_intr -= hz*(tm.tv_sec - btm.tv_sec) +
	    hz*(tm.tv_usec - btm.tv_usec)/1000000;
	stats_all.s3.v_intr -= hz*(tm.tv_sec - btm.tv_sec) +
	    hz*(tm.tv_usec - btm.tv_usec)/1000000;
	stats_all.s2.v_swtch = cnt.v_swtch;
	stats_all.s3.v_swtch = cnt.v_swtch;

    {	FILE *fh;
	if ((fh = fopen("/proc/net/dev", "r")) != NULL) {
	    char buf[512], *s;
	    struct interface myif, *ife = &myif;

	    fgets(buf, sizeof buf, fh);
	    fgets(buf, sizeof buf, fh);
	    procnetdev_vsn = procnetdev_version(buf);

	    while(fgets(buf, sizeof buf, fh)) {
		s = get_name(ife->name, buf);
		get_dev_fields(s, ife);
		stats_all.s1.if_ipackets += ife->stats.rx_packets;
		stats_all.s1.if_opackets += ife->stats.tx_packets;
		stats_all.s1.if_ierrors += ife->stats.rx_errors;
		stats_all.s1.if_oerrors += ife->stats.tx_errors;
		stats_all.s1.if_collisions += ife->stats.collisions;
		stats_all.s2.if_ipackets += ife->stats.rx_packets;
		stats_all.s2.if_opackets += ife->stats.tx_packets;
		stats_all.s2.if_ierrors += ife->stats.rx_errors;
		stats_all.s2.if_oerrors += ife->stats.tx_errors;
		stats_all.s2.if_collisions += ife->stats.collisions;
		stats_all.s3.if_ipackets += ife->stats.rx_packets;
		stats_all.s3.if_opackets += ife->stats.tx_packets;
		stats_all.s3.if_ierrors += ife->stats.rx_errors;
		stats_all.s3.if_oerrors += ife->stats.tx_errors;
		stats_all.s3.if_collisions += ife->stats.collisions;
	    }

	    fclose(fh);
	}
    }

/* ======================================= */

	stats_all.s3.curtime.tv_sec = tm.tv_sec;
	stats_all.s3.curtime.tv_usec = tm.tv_usec;
	alarm(1);
	errno = save_errno;
}

/*
 * returns true if have a disk
 */
static int
havedisk(void)
{
	return dk_ndrive != 0;
}

#else	/* __linux__ */

static void
updatestat(void)
{
	long off;
	int i, save_errno = errno;
	struct vmmeter cnt;
	struct ifnet ifnet;
	double avrun[3];
	struct timeval tm, btm;
#ifdef BSD
	long *cp_time = cur.cp_time;
#endif

#ifdef DEBUG
	syslog(LOG_DEBUG, "entering updatestat");
#endif
	if (sincelastreq >= closedown) {
#ifdef DEBUG
		syslog(LOG_DEBUG, "about to closedown");
#endif
		if (from_inetd)
			exit(0);
		else {
			stat_is_init = 0;
			return;
		}
	}
	sincelastreq++;

	/*
	 * dkreadstats reads in the "disk_count" as well as the "disklist"
	 * statistics.  It also retrieves "hz" and the "cp_time" array.
	 */
	dkreadstats();
	memset(stats_all.s1.dk_xfer, '\0', sizeof(stats_all.s1.dk_xfer));
	for (i = 0; i < dk_ndrive && i < DK_NDRIVE; i++)
		stats_all.s1.dk_xfer[i] = cur.dk_xfer[i];
	
#ifdef BSD
	for (i = 0; i < CPUSTATES; i++)
		stats_all.s1.cp_time[i] = cp_time[cp_xlat[i]];
#else
	if (kvm_read(kfd, (long)nl[X_HZ].n_value, (char *)&hz, sizeof hz) !=
	    sizeof hz) {
		syslog(LOG_ERR, "can't read hz from kmem");
		exit(1);
	}
 	if (kvm_read(kfd, (long)nl[X_CPTIME].n_value,
	    (char *)stats_all.s1.cp_time, sizeof (stats_all.s1.cp_time))
	    != sizeof (stats_all.s1.cp_time)) {
		syslog(LOG_ERR, "can't read cp_time from kmem");
		exit(1);
	}
#endif
#ifdef BSD
	(void)getloadavg(avrun, sizeof(avrun) / sizeof(avrun[0]));
#endif
	stats_all.s2.avenrun[0] = avrun[0] * FSCALE;
	stats_all.s2.avenrun[1] = avrun[1] * FSCALE;
	stats_all.s2.avenrun[2] = avrun[2] * FSCALE;
 	if (kvm_read(kfd, (long)nl[X_BOOTTIME].n_value,
	    (char *)&btm, sizeof (stats_all.s2.boottime))
	    != sizeof (stats_all.s2.boottime)) {
		syslog(LOG_ERR, "can't read boottime from kmem");
		exit(1);
	}
	stats_all.s2.boottime.tv_sec = btm.tv_sec;
	stats_all.s2.boottime.tv_usec = btm.tv_usec;


#ifdef DEBUG
	syslog(LOG_DEBUG, "%d %d %d %d", stats_all.s1.cp_time[0],
	    stats_all.s1.cp_time[1], stats_all.s1.cp_time[2],
	    stats_all.s1.cp_time[3]);
#endif

 	if (kvm_read(kfd, (long)nl[X_CNT].n_value, (char *)&cnt, sizeof cnt) !=
	    sizeof cnt) {
		syslog(LOG_ERR, "can't read cnt from kmem");
		exit(1);
	}
	stats_all.s1.v_pgpgin = cnt.v_pgpgin;
	stats_all.s1.v_pgpgout = cnt.v_pgpgout;
	stats_all.s1.v_pswpin = cnt.v_pswpin;
	stats_all.s1.v_pswpout = cnt.v_pswpout;
	stats_all.s1.v_intr = cnt.v_intr;
	gettimeofday(&tm, (struct timezone *) 0);
	stats_all.s1.v_intr -= hz*(tm.tv_sec - btm.tv_sec) +
	    hz*(tm.tv_usec - btm.tv_usec)/1000000;
	stats_all.s2.v_swtch = cnt.v_swtch;

#ifndef BSD
 	if (kvm_read(kfd, (long)nl[X_DKXFER].n_value,
	    (char *)stats_all.s1.dk_xfer, sizeof (stats_all.s1.dk_xfer))
	    != sizeof (stats_all.s1.dk_xfer)) {
		syslog(LOG_ERR, "can't read dk_xfer from kmem");
		exit(1);
	}
#endif

	stats_all.s1.if_ipackets = 0;
	stats_all.s1.if_opackets = 0;
	stats_all.s1.if_ierrors = 0;
	stats_all.s1.if_oerrors = 0;
	stats_all.s1.if_collisions = 0;
	for (off = (long)ifnetq.tqh_first, i = 0; off && i < numintfs; i++) {
		if (kvm_read(kfd, off, (char *)&ifnet, sizeof ifnet) !=
		    sizeof ifnet) {
			syslog(LOG_ERR, "can't read ifnet from kmem");
			exit(1);
		}
		stats_all.s1.if_ipackets += ifnet.if_data.ifi_ipackets;
		stats_all.s1.if_opackets += ifnet.if_data.ifi_opackets;
		stats_all.s1.if_ierrors += ifnet.if_data.ifi_ierrors;
		stats_all.s1.if_oerrors += ifnet.if_data.ifi_oerrors;
		stats_all.s1.if_collisions += ifnet.if_data.ifi_collisions;
		off = (long)ifnet.if_list.tqe_next;
	}
	gettimeofday((struct timeval *)&stats_all.s3.curtime,
		(struct timezone *) 0);
	alarm(1);
	errno = save_errno;
}

static void
setup(void)
{
	struct ifnet ifnet;
	long off;
	char errbuf[_POSIX2_LINE_MAX];

	kfd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf);
	if (kfd == NULL) {
		syslog(LOG_ERR, "%s", errbuf);
		exit (1);
	}

	if (kvm_nlist(kfd, nl) != 0) {
		syslog(LOG_ERR, "can't get namelist");
		exit (1);
	}

	if (kvm_read(kfd, (long)nl[X_IFNET].n_value, &ifnetq,
	    sizeof ifnetq) != sizeof ifnetq) {
		syslog(LOG_ERR, "can't read ifnet queue head from kmem");
		exit(1);
	}

	numintfs = 0;
	for (off = (long)ifnetq.tqh_first; off;) {
		if (kvm_read(kfd, off, (char *)&ifnet, sizeof ifnet) !=
		    sizeof ifnet) {
			syslog(LOG_ERR, "can't read ifnet from kmem");
			exit(1);
		}
		numintfs++;
		off = (long)ifnet.if_list.tqe_next;
	}
#ifdef BSD
	dkinit(0);
#endif
}

/*
 * returns true if have a disk
 */
static int
havedisk(void)
{
#ifdef BSD
	return dk_ndrive != 0;
#else
	int i, cnt;
	long  xfer[DK_NDRIVE];

	if (kvm_nlist(kfd, nl) != 0) {
		syslog(LOG_ERR, "can't get namelist");
		exit (1);
	}

	if (kvm_read(kfd, (long)nl[X_DKXFER].n_value,
		     (char *)xfer, sizeof xfer) != sizeof xfer) {
		syslog(LOG_ERR, "can't read dk_xfer from kmem");
		exit(1);
	}
	cnt = 0;
	for (i=0; i < DK_NDRIVE; i++)
		cnt += xfer[i];
	return (cnt != 0);
#endif
}
#endif	/* __linux__ */

void
rstat_service(struct svc_req *rqstp, SVCXPRT *transp)
{
	union {
		int fill;
	} argument;
	char *result;
	xdrproc_t xdr_argument, xdr_result;
	typedef char *(*localproc_t) __P((void *, struct svc_req *));
	localproc_t local;

	switch (rqstp->rq_proc) {
	case NULLPROC:
		(void)svc_sendreply(transp, xdr_void, (char *)NULL);
		return;
		break;

	case RSTATPROC_STATS:
		xdr_argument = (xdrproc_t)xdr_void;
		xdr_result = (xdrproc_t)xdr_statstime;
		switch (rqstp->rq_vers) {
		case RSTATVERS_ORIG:
			local = (localproc_t) rstatproc_stats_1_svc;
			break;
		case RSTATVERS_SWTCH:
			local = (localproc_t) rstatproc_stats_2_svc;
			break;
		case RSTATVERS_TIME:
			local = (localproc_t) rstatproc_stats_3_svc;
			break;
		default:
			svcerr_progvers(transp, RSTATVERS_ORIG, RSTATVERS_TIME);
			return;
			break;
		}
		break;

	case RSTATPROC_HAVEDISK:
		xdr_argument = (xdrproc_t)xdr_void;
		xdr_result = (xdrproc_t)xdr_u_int;
		switch (rqstp->rq_vers) {
		case RSTATVERS_ORIG:
			local = (localproc_t) rstatproc_havedisk_1_svc;
			break;
		case RSTATVERS_SWTCH:
			local = (localproc_t) rstatproc_havedisk_2_svc;
			break;
		case RSTATVERS_TIME:
			local = (localproc_t) rstatproc_havedisk_3_svc;
			break;
		default:
			svcerr_progvers(transp, RSTATVERS_ORIG, RSTATVERS_TIME);
			return;
			break;
		}
		break;

	default:
		svcerr_noproc(transp);
		return;
		break;
	}
	memset(&argument, 0, sizeof(argument));
	if (!svc_getargs(transp, xdr_argument, (caddr_t)&argument)) {
		/* try to free already allocated memory during decoding */
		svc_freeargs(transp, xdr_argument, (caddr_t)&argument);
		svcerr_decode(transp);
		return;
	}
	result = (*local)(&argument, rqstp);
	if (result != NULL && !svc_sendreply(transp, xdr_result, result)) {
		svcerr_systemerr(transp);
	}
	if (!svc_freeargs(transp, xdr_argument, (caddr_t)&argument)) {
		syslog(LOG_ERR, "unable to free arguments");
		exit(1);
	}
}
